// Unreal Fighting Framework by Kevin Poretti

#include "FFStateMachineComponent.h"

// FF includes
#include "FFState.h"

UFFStateMachineComponent::UFFStateMachineComponent()
{
    // Don't use Unreal's tick instead use a fixed tick
	PrimaryComponentTick.bCanEverTick = false;
}


void UFFStateMachineComponent::Initialize()
{
    for(const TSubclassOf<UFFStateBehavior>& CurrState : DefaultStates)
    {
        UFFStateBehavior* TempState = AddState(CurrState);
        if(!CurrentState) // first state to be created is the entry into this state machine
        {
            CurrentState = TempState;
            CurrentState->Enter(GetCurrentStateContext());
        }
    }
}


void UFFStateMachineComponent::InitActorInfo(AActor* InOwner, AActor* InAvatar)
{
    Owner = InOwner;
	Avatar = InAvatar;
}


UFFStateBehavior* UFFStateMachineComponent::AddState(TSubclassOf<UFFStateBehavior> StateClassToAdd)
{
    UFFStateBehavior* TempState = NewObject<UFFStateBehavior>(this, StateClassToAdd);
    if(TempState)
    {
        States.Add(TempState);
        return TempState;
    }

    return nullptr;
}


void UFFStateMachineComponent::AddStates(const TArray<TSubclassOf<UFFStateBehavior>>& StateClassesToAdd)
{
    for(const TSubclassOf<UFFStateBehavior>& CurrState : StateClassesToAdd)
    {
        AddState(CurrState);
    }
}


void UFFStateMachineComponent::RemoveState(FName StateToRemove)
{
    UE_LOG(LogTemp, Error, TEXT("UFFStateMachineComponent::RemoveState is not yet implemented"));
}

void UFFStateMachineComponent::SwitchStates(UFFStateBehavior* NewState)
{
    check(NewState);


    CurrentState->Exit(GetCurrentStateContext());
    CurrentState = NewState;
    CurrentState->Enter(GetCurrentStateContext());
    TicksInState = 0;
}


FName UFFStateMachineComponent::GetCurrentStateName() const
{
    return CurrentState ? CurrentState->Name : NAME_None;
}


FFFStateContext UFFStateMachineComponent::GetCurrentStateContext() const
{
    FFFStateContext CurrStateContext;
    CurrStateContext.Owner = Owner;
    CurrStateContext.Avatar = Avatar;
    return CurrStateContext;
}


UFFStateBehavior* UFFStateMachineComponent::FindStateWithName(FName StateName)
{
    for (UFFStateBehavior* CurrState : States)
    {
        if(CurrState->Name == StateName)
        {
            return CurrState;
        }
    }

	UE_LOG(LogTemp, Warning,
			TEXT("Could not find state in state machine with name %s on %s"), *StateName.ToString(), *Owner->GetName());

	return nullptr;
}


void UFFStateMachineComponent::BeginPlay()
{
	Super::BeginPlay();

    Initialize();
}


void UFFStateMachineComponent::FixedTick(float OneFrame)
{
    // Should we switch states?

    for(UFFStateBehavior* CurrState : States)
    {
        // Check if the state is enabled

        // Check state entry conditions if there are any

        // Check input conditions if there are any

        // If all state entry conditions are good and at least one input condition is good then we can transition

        // Lastly just check if the state we're about to transition into isn't the current state.
        // It is OK to transition if state's "CanTransitionToSelf" is true

        // SwitchStates(NewState);
        // return;
    }

    // CurrentState should never be null
    // TODO: Should probably assert or whatever UE's equivalent is
    check(CurrentState);

    // Tick current state
    TicksInState++;
    CurrentState->Update(OneFrame, GetCurrentStateContext());

    // Debug
}

